@@ -1,20 +1,22 @@
 import asyncio
+import copy
 import enum
 import ipaddress
 import logging
 import random
 import secrets
 import socket
+import threading
 from itertools import count
 from typing import Dict, List, Optional, Set, Text, Tuple, Union, cast
 
 import netifaces
 
-from . import stun, turn
+from . import mdns, stun, turn
 from .candidate import Candidate, candidate_foundation, candidate_priority
 from .utils import random_string
 
-logger = logging.getLogger("ice")
+logger = logging.getLogger(__name__)
 
 ICE_COMPLETED = 1
 ICE_FAILED = 2
@@ -25,6 +27,18 @@ CONSENT_INTERVAL = 5
 connection_id = count()
 protocol_id = count()
 
+_mdns = threading.local()
+
+
+async def get_or_create_mdns_protocol() -> mdns.MDnsProtocol:
+    if not hasattr(_mdns, "lock"):
+        _mdns.lock = asyncio.Lock()
+        _mdns.protocol = None
+    async with _mdns.lock:
+        if _mdns.protocol is None:
+            _mdns.protocol = await mdns.create_mdns_protocol()
+    return _mdns.protocol
+
 
 def candidate_pair_priority(
     local: Candidate, remote: Candidate, ice_controlling: bool
@@ -102,8 +116,6 @@ def sort_candidate_pairs(pairs, ice_controlling: bool) -> None:
 def validate_remote_candidate(candidate: Candidate) -> Candidate:
     """
     Check the remote candidate is supported.
-
-    mDNS candidates are not supported yet.
     """
     if candidate.type not in ["host", "relay", "srflx"]:
         raise ValueError('Unexpected candidate type "%s"' % candidate.type)
@@ -236,6 +248,14 @@ class StunProtocol(asyncio.DatagramProtocol):
         return "protocol(%s)" % self.id
 
 
+class ConnectionEvent:
+    pass
+
+
+class ConnectionClosed(ConnectionEvent):
+    pass
+
+
 class Connection:
     """
     An ICE connection for a single media stream.
@@ -285,6 +305,7 @@ class Connection:
         self.turn_transport = turn_transport
 
         # private
+        self._closed = False
         self._components = set(range(1, components + 1))
         self._check_list: List[CandidatePair] = []
         self._check_list_done = False
@@ -292,6 +313,7 @@ class Connection:
         self._early_checks: List[
             Tuple[stun.Message, Tuple[str, int], StunProtocol]
         ] = []
+        self._event_waiter: Optional[asyncio.Future[ConnectionEvent]] = None
         self._id = next(connection_id)
         self._local_candidates: List[Candidate] = []
         self._local_candidates_end = False
@@ -339,11 +361,31 @@ class Connection:
             self._remote_candidates_end = True
             return
 
+        # resolve mDNS candidate
+        if mdns.is_mdns_hostname(remote_candidate.host):
+            mdns_protocol = await get_or_create_mdns_protocol()
+            remote_addr = await mdns_protocol.resolve(remote_candidate.host)
+            if remote_addr is None:
+                self.__log_info(
+                    f'Remote candidate "{remote_candidate.host}" could not be resolved'
+                )
+                return
+            self.__log_info(
+                f'Remote candidate "{remote_candidate.host}" resolved to {remote_addr}'
+            )
+
+            copy_candidate = copy.copy(remote_candidate)
+            copy_candidate.host = remote_addr
+            await self.add_remote_candidate(copy_candidate)
+            return
+
         # validate the remote candidate
         try:
             validate_remote_candidate(remote_candidate)
         except ValueError as e:
-            self.__log_info(f"Candidate not valid: {e}")
+            self.__log_info(
+                f'Remote candidate "{remote_candidate.host}" is not valid: {e}'
+            )
             return
         self._remote_candidates.append(remote_candidate)
 
@@ -367,10 +409,12 @@ class Connection:
             addresses = get_host_addresses(
                 use_ipv4=self._use_ipv4, use_ipv6=self._use_ipv6
             )
-            for component in self._components:
-                self._local_candidates += await self.get_component_candidates(
-                    component=component, addresses=addresses
-                )
+            coros = [
+                self.get_component_candidates(component=component, addresses=addresses)
+                for component in self._components
+            ]
+            for candidates in await asyncio.gather(*coros):
+                self._local_candidates += candidates
             self._local_candidates_end = True
 
     def get_default_candidate(self, component: int) -> Optional[Candidate]:
@@ -459,6 +503,25 @@ class Connection:
         self._protocols.clear()
         self._local_candidates.clear()
 
+        # emit event
+        if not self._closed:
+            self._emit_event(ConnectionClosed())
+            self._closed = True
+
+    async def get_event(self) -> Optional[ConnectionEvent]:
+        """
+        Return the next `ConnectionEvent` or `None` if the connection is
+        already closed.
+
+        This method may only be called once at a time.
+        """
+        assert self._event_waiter is None, "already awaiting event"
+        if self._closed:
+            return None
+        loop = asyncio.get_event_loop()
+        self._event_waiter = loop.create_future()
+        return await asyncio.shield(self._event_waiter)
+
     async def recv(self) -> bytes:
         """
         Receive the next datagram.
@@ -748,6 +811,12 @@ class Connection:
         self.__log_info("Check %s %s -> %s", pair, pair.state, state)
         pair.state = state
 
+    def _emit_event(self, event: ConnectionEvent) -> None:
+        if self._event_waiter is not None:
+            waiter = self._event_waiter
+            self._event_waiter = None
+            waiter.set_result(event)
+
     def _find_pair(
         self, protocol: StunProtocol, remote_candidate: Candidate
     ) -> Optional[CandidatePair]:
@@ -763,8 +832,10 @@ class Connection:
         self, component: int, addresses: List[str], timeout: int = 5
     ) -> List[Candidate]:
         candidates = []
-
         loop = asyncio.get_event_loop()
+
+        # gather host candidates
+        host_protocols = []
         for address in addresses:
             # create transport
             try:
@@ -775,7 +846,7 @@ class Connection:
                 self.__log_info("Could not bind to %s - %s", address, exc)
                 continue
             protocol = cast(StunProtocol, protocol)
-            self._protocols.append(protocol)
+            host_protocols.append(protocol)
 
             # add host candidate
             candidate_address = protocol.transport.get_extra_info("sockname")
@@ -789,11 +860,12 @@ class Connection:
                 type="host",
             )
             candidates.append(protocol.local_candidate)
+        self._protocols += host_protocols
 
         # query STUN server for server-reflexive candidates (IPv4 only)
         if self.stun_server:
             tasks = []
-            for protocol in self._protocols:
+            for protocol in host_protocols:
                 if ipaddress.ip_address(protocol.local_candidate.host).version == 4:
                     tasks.append(
                         asyncio.ensure_future(
